;; -*- lexical-binding: t -*-

(require 'cl-lib)
(require 'filenotify)
(require 'tramp)
(require 'tramp-sh)
(require 'seq)

(defvar checkfn-tests nil)

(defvar checkfn-rationalize-results t
  "Whether to map filenames to symbolic names.")

(defvar checkfn-timeout 12
  "Timeout for a single test.")

(defvar checkfn-watch-flags '(changed attribute-change)
  "Flags used for `file-notify-add-watch'.")

(defun checkfn-delete-files (filenames)
  (dolist (filename filenames)
    (when (and (stringp filename)
               (file-exists-p filename))
      (if (file-directory-p filename)
          (delete-directory filename t)
        (delete-file filename)))))

(defmacro checkfn-gather-events-let* (bindings &rest body)
  "Watch files bound in BINDINGS collecting all events.

BINDINGS should bind variables to filenames.  If BODY does not
delete these filenames, this macro will do it.

Return alist with elements (FILENAME . EVENTS)."
  (declare (indent 1) (debug ((&rest (symbolp form)) body)))
  (let ((watches (make-symbol "watches"))
        (results (make-symbol "results"))
        (filenames (make-symbol "filenames"))
        (filealist (make-symbol "filealist"))
        (done (make-symbol "done")))
    `(let* (,@bindings
            (,filenames (list ,@(mapcar #'car bindings)))
            (,filealist (list ,@(mapcar (lambda (elt)
                                          `(cons ,(car elt) ',(car elt)))
                                        bindings)))
            (,results (mapcar #'list ,filenames))
            (,done (mapcar #'list ,filenames))
            (,watches (progn
                        (mapcar
                         (lambda (filename)
                           (file-notify-add-watch
                            filename
                            checkfn-watch-flags
                            (lambda (event)
                              (when (eq (nth 1 event) 'stopped)
                                (let ((elt (assoc filename ,done)))
                                  (setcdr elt t)))
                              (let ((elt (assoc filename ,results)))
                                (setcdr elt (append (cdr elt) (list (cdr event))))))))
                         ,filenames))))
       (unwind-protect
           (progn
             (progn ,@body)
             (checkfn-delete-files ,filenames)
             (with-timeout
                 (checkfn-timeout
                  (dolist (elt ,results)
                    (unless (cdr (assoc (car elt) ,done))
                      (setcdr elt (append (cdr elt)
                                          `((timedout ,checkfn-timeout))))))
                  (dolist (watch ,watches)
                    (ignore-errors
                      (file-notify-rm-watch watch))))
               (while (memq nil (mapcar #'cdr ,done))
                 (read-event nil nil 0.1)))
             (if checkfn-rationalize-results
                 (checkfn-gather-events-rationalize-results ,results ,filealist)
               ,results))
         (dolist (watch ,watches)
           (ignore-errors
             (file-notify-rm-watch watch)))))))

(defun checkfn-gather-events-rationalize-results (results alist)
  (let ((nanons 0))
    (cl-labels ((aget (key)
                  (let ((elt (assoc key alist)))
                    (if elt
                        (cdr elt)
                      (let ((anon (intern (format "anon-%d" (cl-incf nanons)))))
                        (push (cons key anon) alist)
                        anon)))))
      (mapcar (pcase-lambda (`(,filename . ,events))
                (cons (aget filename)
                      (mapcar (pcase-lambda (`(,action ,file . ,rest))
                                (if (eq action 'timedout)
                                    (list action file)
                                  (if rest
                                      `(,action ,(aget file)
                                        ,(aget (car rest)))
                                    `(,action ,(aget file)))))
                              events)))
              results))))


(defmacro define-checkfn-test (name &optional documentation &rest body)
  (cl-check-type name symbol)
  (let ((fn (intern (format "checkfn-test-%s" name))))
    `(progn
       (defun ,fn () ,documentation ,@body)
       (setq checkfn-tests
             (append (cl-remove ',name checkfn-tests :key #'car)
                     (list (cons ',name ',fn))))
       ',name)))

(defun checkfn-chattr (filename)
  (cl-assert (file-exists-p filename))
  (set-file-times
   filename
   (seconds-to-time (1+ (time-to-seconds (current-time))))))

(defun checkfn-create (filename &optional directory-p)
  (cl-assert (not (file-exists-p filename)))
  (if directory-p
      (make-directory filename)
    (with-temp-file filename)))

(defun checkfn-make-file (&optional dir directory-p)
  (cl-check-type dir (or null string))
  (let ((temporary-file-directory
         (or dir temporary-file-directory)))
    (make-temp-file "checkfn" directory-p)))

(defun checkfn-make-name (&optional dir)
  (cl-check-type dir (or null string))
  (make-temp-name (format "%s/checkfn"
                          (directory-file-name
                           (or dir temporary-file-directory)))))


(defmacro checkfn-cleanup-files-let* (bindings &rest body)
  (declare (indent 1) (debug ((&rest (symbolp form)) body)))
  (let ((files (make-symbol "files")))
    `(let* (,@bindings
            (,files (list ,@(mapcar #'car bindings))))
       (unwind-protect
           (progn ,@body)
         (checkfn-delete-files ,files)))))


;; +===================================================================================+
;; | Regular Files
;; +===================================================================================+

(define-checkfn-test file-create
  (if (eq (checkfn-backend) 'kqueue)
      'skipped
    (checkfn-gather-events-let* ((file (checkfn-make-name)))
      (checkfn-create file))))

(define-checkfn-test file-read
  (checkfn-gather-events-let* ((file (checkfn-make-file)))
    (with-temp-buffer
      (insert-file-contents file))))

(define-checkfn-test file-write
  (checkfn-gather-events-let* ((file (checkfn-make-file)))
    (with-temp-file file
      (insert "XXX"))))

(define-checkfn-test file-attrib
  (checkfn-gather-events-let* ((file (checkfn-make-file)))
    (checkfn-chattr file)
    (delete-file file)))

(define-checkfn-test file-rename
  (checkfn-gather-events-let* ((dir (checkfn-make-file nil t))
                               (dir/src (checkfn-make-file dir))
                               (dir/dest (checkfn-make-file dir)))
    (rename-file dir/src dir/dest t)))


;; +===================================================================================+
;; | Directories
;; +===================================================================================+


(define-checkfn-test dir-create
  (if (eq (checkfn-backend) 'kqueue)
      'skipped
    (checkfn-gather-events-let* ((dir (checkfn-make-name)))
      (checkfn-create dir t))))

(define-checkfn-test dir-read
  (checkfn-gather-events-let* ((dir (checkfn-make-file nil t)))
    (directory-files dir)))

(define-checkfn-test dir-create-file
  (checkfn-gather-events-let* ((dir (checkfn-make-file nil t))
                               (dir/file (checkfn-make-file dir)))))

(define-checkfn-test dir-create-dir
  (checkfn-gather-events-let* ((dir (checkfn-make-file nil t))
                               (dir/dir (checkfn-make-file dir t)))))

(define-checkfn-test dir-attrib
  (checkfn-gather-events-let* ((dir (checkfn-make-file nil t)))
    (checkfn-chattr dir)))

(define-checkfn-test dir-rename
  (if (eq (checkfn-backend) 'kqueue)
      'skipped
    (checkfn-gather-events-let* ((dir (checkfn-make-file nil t))
                                 (dir/src (checkfn-make-file dir t))
                                 (dir/dest (checkfn-make-name)))
      (rename-file dir/src dir/dest t))))


;; +===================================================================================+
;; | Runner
;; +===================================================================================+

(defun checkfn-backend (&optional filename)
  (unless filename (setq filename temporary-file-directory))
  (let* ((watch (file-notify-add-watch
                 filename
                 '(change attribute-change)
                 #'ignore)))
    (unwind-protect
        (cond
         ((processp watch)
          (intern
           (file-name-sans-extension
            (replace-regexp-in-string
             "<[[:digit:]]+>" ""
             (process-name watch)))))
         (t file-notify--library))
      (ignore-errors (file-notify-rm-watch watch)))))

(defun checkfn-run (&optional tests)
  (unless tests
    (setq tests checkfn-tests))
  (let (results)
    (dolist (elt tests)
      (when (symbolp elt)
        (setq elt (assq elt checkfn-tests)))
      (cl-check-type elt cons)
      (cl-check-type (car elt) symbol)
      (cl-check-type (cdr elt) function)
      (message "Test: %s" (car elt))
      (checkfn-cleanup-files-let*  ((temporary-file-directory
                                     (checkfn-make-file nil t)))
        (pcase-let ((`(,name . ,fn) elt))
          (push (cons name (condition-case err
                               (funcall fn)
                             (error
                              (message "Error: %s" (error-message-string err))
                              'failed)))
                results))))
    (cons (checkfn-backend)
          (nreverse results))))

(defun checkfn-run-pp (&optional outfile tests)
  (let ((print-length nil)
        (print-level nil)
        (print-quoted t)
        (standard-output
         (if outfile (with-current-buffer (find-file-noselect outfile)
                       (erase-buffer)
                       (current-buffer))
           standard-output)))
    (message "Backend: %s" (checkfn-backend))
    (when outfile
      (message "Writing to `%s'" outfile))
    (pp (checkfn-run tests))
    (when outfile
      (with-current-buffer standard-output
        (save-buffer)))))

(defmacro checkfn-with-tramp-mock-method (&rest body)
  `(let ((tramp-methods
          (cons '("mock"
                  (tramp-login-program        "sh")
                  (tramp-login-args           (("-i")))
                  (tramp-remote-shell         "/bin/sh")
                  (tramp-remote-shell-args    ("-c"))
                  (tramp-connection-timeout   10))
                tramp-methods)))
     ,@body))

(defun checkfn-gfilemonitor-run-pp (&optional outfile tests)
  (cl-letf (((symbol-function 'tramp-get-remote-inotifywait)
             #'ignore))
    (checkfn-with-tramp-mock-method
     (let ((temporary-file-directory
            (concat "/mock::" temporary-file-directory))
           (tramp-verbose 0))
       (checkfn-run-pp outfile tests)))))

(defun checkfn-inotifywait-run-pp (&optional outfile tests)
  (cl-letf (((symbol-function 'tramp-get-remote-gvfs-monitor-dir)
             #'ignore))
    (checkfn-with-tramp-mock-method
     (let ((temporary-file-directory
            (concat "/mock::" temporary-file-directory))
           (tramp-verbose 0))
       (checkfn-run-pp outfile tests)))))

(defun checkfn-run-batch (&optional outdir backend)
  (let ((outfile (expand-file-name
                  (format "checkfn-%s-log.el"
                          (or backend (checkfn-backend)))
                  outdir)))
    (message "%S" backend)
    (pcase backend
      ('gfilemonitor
       (checkfn-gfilemonitor-run-pp outfile))
      ('inotifywait
       (checkfn-inotifywait-run-pp outfile))
      ((guard (or (null backend)
                  (eq backend (checkfn-backend))))
       (checkfn-run-pp outfile))
      (_ (error "Backend not available: %s" backend)))))




;; +===================================================================================+
;; | Table
;; +===================================================================================+

(defun checkfn-table-draw-test-result (test-results c1 c2 c3)
  (let ((name (caar test-results))
        (actions
         (let ((watches (seq-some (lambda (elt)
                                    (and (consp elt)
                                         (mapcar #'car elt)))
                                  (mapcar #'cdr test-results))))
           (mapcar (lambda (events)
                     (if (not (consp events))
                         (mapcar (lambda (watch)
                                   (cons watch (list events)))
                                 (or watches (list "")))
                       events))
                   (mapcar #'cdr test-results)))))
    (princ name)
    (move-to-column c1 t)
    (apply
     #'cl-mapcar
     (lambda (&rest watches)
       (princ (caar watches))
       (move-to-column (+ c1 c2) t)
       (let ((actions (mapcar #'cdr watches)))
         (while (not (seq-every-p #'null actions))
           (let ((column (+ c1 c2)))
             (setq actions
                   (mapcar
                    (lambda (action)
                      (prog1
                          (when action
                            (if (consp (car action))
                                (dolist (action (car action))
                                  (princ
                                   (cl-case action
                                     (attribute-changed 'attr-changed)
                                     (t action)))
                                  (princ " "))
                              (princ (car action)))
                            (cdr action))
                        (cl-incf column c3)
                        (move-to-column column t)))
                    actions)))
           (terpri)
           (move-to-column (+ c1 c2) t))
         (terpri)
         (move-to-column c1 t)))
     actions)
    (move-to-column 0 t)))

(defun checkfn-make-table (&rest results)
  (setq results (mapcar (lambda (result)
                          (if (stringp result)
                              (with-temp-buffer
                                (insert-file-contents result)
                                (goto-char 1)
                                (read (current-buffer)))
                            result))
                        results))
  (with-current-buffer (get-buffer-create "*checkfn-table*")
    (let* ((inhibit-read-only t)
           (standard-output (current-buffer))
           (backends (mapcar #'car results))
           (columns (list 16 12 32)))
      (cl-destructuring-bind (c1 c2 c3)
          columns
        (cl-labels ((hline ()
                      (princ (make-string (+ c1 c2 (* c3 (length backends))) ?-))
                      (terpri)))
          (erase-buffer)
          (setq-local truncate-lines t)
          (let ((column (+ c1 c2)))
            (move-to-column column t)
            (dolist (backend backends)
              (pp backend)
              (cl-incf column c3)
              (move-to-column column t))
            (terpri))
          (hline)
          (dotimes (n (length (cdr (car results))))
            (checkfn-table-draw-test-result
             (checkfn-table-nth-test results n)
             c1 c2 c3)
            (hline))
          (pop-to-buffer (current-buffer))
          (goto-char (point-min))
          (delete-other-windows))))))
